@@ -315,7 +315,7 @@ _String*            _CalcNode::GetBranchSpec (void) {
 
 //__________________________________________________________________________________
 
-void _CalcNode::ReplaceModel (_String& modelName, _VariableContainer* parent_tree) {
+void _CalcNode::ReplaceModel (_String& modelName, _VariableContainer* parent_tree, _AVLListXL* aCache, bool clean_locals) {
     // TODO: SLKP 20171203 this is FUGLY
   RemoveModel    ();
 
@@ -337,8 +337,10 @@ void _CalcNode::ReplaceModel (_String& modelName, _VariableContainer* parent_tre
 
   long my_var_index = theIndex;
 
-  DeleteVariable (theIndex, false); // this will clean up all the node.xx variables
-  InitializeCN   (modelName, 0 , parent_tree);
+  if (clean_locals) {
+      DeleteVariable (theIndex, false); // this will clean up all the node.xx variables
+  }
+  InitializeCN   (modelName, 0 , parent_tree, aCache);
 
   if (container_object) {
     _Variable * new_node = LocateVar(my_var_index);
@@ -442,9 +444,9 @@ void    _CalcNode::Duplicate (BaseRefConst theO) {
 
 //_______________________________________________________________________________________________
 
-bool        _CalcNode::HasChanged(bool) {
+bool        _CalcNode::HasChanged(bool, _AVLListX * cache) {
     
-    return _VariableContainer::HasChanged() || categoryVariables.Any([&] (long cat_idx, unsigned long) -> bool {
+    return _VariableContainer::HasChanged(false, cache) || categoryVariables.Any([&] (long cat_idx, unsigned long) -> bool {
         return LocateVar (cat_idx)->HasChanged();
     });
  }
@@ -452,36 +454,67 @@ bool        _CalcNode::HasChanged(bool) {
 
 //_______________________________________________________________________________________________
 
-bool        _CalcNode::NeedNewCategoryExponential(long catID) const {
+bool        _CalcNode::NeedNewCategoryExponential(long catID, _AVLListX * cache) const {
+    
 
-    if (_VariableContainer::NeedToExponentiate(catID>=0)) {
+    auto handle_cache = [&](long cat_idx, long idx, auto checker) -> bool {
+        if (cache) {
+            long pre_comp = cache->GetDataByKey(cat_idx);
+            if (pre_comp >= 0) {
+                return pre_comp;
+            }
+        }
+        bool has_changed = checker (cat_idx, idx);
+        if (cache) {
+            long ins_idx = cache->InsertNumber(cat_idx);
+            cache->SetXtra(ins_idx, has_changed);
+        }
+        return has_changed;
+    };
+    
+    auto cat_overall = [](long cat_idx, long) -> bool {
+        return LocateVar (cat_idx)->HasChanged();
+    };
+    
+    auto cat_class = [&](long cat_idx, long i) -> bool {
+        return ((_CategoryVariable*)LocateVar (cat_idx))->HaveParametersChanged(remapMyCategories.list_data[catID*(categoryVariables.countitems()+1)+i+1]);
+    };
+    
+    if (_VariableContainer::NeedToExponentiate(catID>=0, cache)) {
         return true;
     }
 
     if (catID==-1) {
         return !compExp || categoryVariables.Any([&] (long cat_idx, unsigned long) -> bool {
-            return LocateVar (cat_idx)->HasChanged();
+            return handle_cache (cat_idx, 0, cat_overall);
         });
     } else {
         return !GetCompExp(catID) || categoryVariables.Any([&] (long cat_idx, unsigned long i) -> bool {
-            return ((_CategoryVariable*)LocateVar (cat_idx))->HaveParametersChanged(remapMyCategories.list_data[catID*(categoryVariables.countitems()+1)+i+1]);
+            return handle_cache (cat_idx, i, cat_class);
         });
     }
     return false;
 }
 
 //_______________________________________________________________________________________________
-bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix* storeRateMatrix, _List* queue, _SimpleList* tags, _List* bufferedOps)
+bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix* storeRateMatrix, _List* queue, _SimpleList* tags, _List* bufferedOps, bool direct_copy)
 {
     // assumed that NeedToExponentiate was called prior to this function
 
     //_Variable* curVar, *locVar;
+    bool    isExplicitForm  = HasExplicitFormModel ();
+
+    if (direct_copy && bufferedOps && bufferedOps->countitems() == 1 && isExplicitForm) {
+        SetCompExp ((_Matrix*)bufferedOps->GetItem(0)->makeDynamic(), totalCategs>1?categID:-1);
+        return false;
+    }
     
     _SimpleList * var_lists [2] = {iVariables, dVariables};
     for (_SimpleList* iterable : var_lists) {
         ForEachLocalVariable(iterable,CopyModelParameterValue);
     }
   
+    /*
     #ifdef _UBER_VERBOSE_MX_UPDATE_DUMP
      // if (1|| likeFuncEvalCallCount == _UBER_VERBOSE_MX_UPDATE_DUMP_LF_EVAL && gVariables) {
         for (unsigned long i=0; i<gVariables->lLength; i++) {
@@ -490,6 +523,7 @@ bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix
         }
       //}
     #endif
+    */
 
     /*
     for (unsigned long i=0; i<categoryVariables.lLength; i++) {
@@ -503,11 +537,11 @@ bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix
 
     if (!storeRateMatrix) {
       if (totalCategs>1) {
-  #ifdef _UBER_VERBOSE_MX_UPDATE_DUMP
-        fprintf (stderr, "[_CalcNode::RecomputeMatrix] Deleting category %ld for node %s at %p\n", categID, GetName()->get_str(), GetCompExp(categID));
-  #endif
-        if (clear_exponentials()) {
-            DeleteObject(GetCompExp(categID, true));
+         if (clear_exponentials()) {
+            #ifdef _UBER_VERBOSE_MX_UPDATE_DUMP
+                  fprintf (stderr, "[_CalcNode::RecomputeMatrix] Deleting category %ld for node %s at %p\n", categID, GetName()->get_str(), GetCompExp(categID));
+            #endif
+           DeleteObject(GetCompExp(categID, true));
         }
       } else {
         if (compExp) {
@@ -518,20 +552,20 @@ bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix
       }
     }
 
-    bool    isExplicitForm  = HasExplicitFormModel ();
-
+ 
     if (isExplicitForm && bufferedOps) {
-        _Matrix * bufferedExp = (_Matrix*)GetExplicitFormModel()->Compute (0,nil, bufferedOps);
-        #ifdef _UBER_VERBOSE_MX_UPDATE_DUMP
+        _Matrix * bufferedExp = (_Matrix*)GetExplicitFormModel(map_global_to_local_category (categID))->Compute (0,nil, bufferedOps);
+         #ifdef _UBER_VERBOSE_MX_UPDATE_DUMP
             fprintf (stderr, "[_CalcNode::RecomputeMatrix] Setting (buffered) category %ld/%ld for node %s\n", categID, totalCategs, GetName()->get_str());
+            ObjectToConsole(bufferedExp);
          #endif
         SetCompExp ((_Matrix*)bufferedExp->makeDynamic(), totalCategs>1?categID:-1);
         return false;
     }
 
     unsigned long      previous_length = queue && tags ? queue->lLength: 0;
-    _Matrix * myModelMatrix = GetModelMatrix(queue,tags);
-
+    _Matrix * myModelMatrix = GetModelMatrix(queue,tags,map_global_to_local_category(categID));
+    
     if (isExplicitForm && !myModelMatrix) { // matrix exponentiations got cached
         if (queue && queue->lLength > previous_length) {
             return true;
@@ -541,10 +575,10 @@ bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix
         }
     } else {
 
-        if (myModelMatrix->MatrixType()!=_POLYNOMIAL_TYPE) {
+        if (!myModelMatrix->is_polynomial()) {
             _Matrix *temp = nil;
             if (isExplicitForm) {
-                temp = (_Matrix*)myModelMatrix->makeDynamic();
+                temp = (_Matrix*)myModelMatrix;
             } else {
                 temp = (_Matrix*)myModelMatrix->MultByFreqs(theModel, true);
             }
@@ -562,6 +596,23 @@ bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix
                 }
             });
             
+            /*
+            _String check ("simulator.substitution_model_Q_component_1");
+            _Variable * check_me = FetchVar (LocateVarByName (check), MATRIX);
+            
+            if (check_me) {
+                StringToConsole("\n\n");
+                _Matrix * m = (_Matrix*)((_Matrix*)((_Matrix*)check_me->GetValue())->ComputeNumeric(true));
+                char nucs[] = {'A','C','G','T'};
+                for (long k = 0; k < 48; k ++) {
+                    for (long k2 = k+1; k2 < 48; k2++) {
+                         if ((*m)(k,k2) != 0.0) {
+                            printf ("%ld %ld (%d) %g\n", k, k2, (nucs[k/16]!=nucs[k2/16]) + (nucs[k%16/4] != nucs[k2%16/4]) + (nucs[k2%4]!=nucs[k%4]), (*m)(k,k2));
+                        }
+                    }
+                }
+            }*/
+            
 
             if (storeRateMatrix) {
                 storeRateMatrix->Duplicate(temp);
@@ -570,9 +621,10 @@ bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix
 
 
             if (queue) {
+                
                 (*queue) << temp;
                 if (tags) {
-                    (*tags) << (isExplicitForm);
+                    (*tags) << (isExplicitForm ? (queue->countitems() - previous_length == 1 ? -1 : 1) : 0);
                 }
                 return isExplicitForm;
             }
@@ -580,7 +632,14 @@ bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix
             #ifdef _UBER_VERBOSE_MX_UPDATE_DUMP
                 fprintf (stderr, "[_CalcNode::RecomputeMatrix] Setting category %ld/%ld for node %s\n", categID, totalCategs, GetName()->get_str());
             #endif
-            SetCompExp ((_Matrix*)(isExplicitForm?temp:temp->Exponentiate(1., true)), totalCategs>1?categID:-1);
+            
+            if (isExplicitForm) {
+                temp->AddAReference();
+                SetCompExp (temp, totalCategs>1?categID:-1);
+            } else {
+                SetCompExp (temp->Exponentiate(1., true), totalCategs>1?categID:-1);
+            }
+            
 
         } else {
             compExp = (_Matrix*)myModelMatrix->Evaluate(false);
@@ -588,6 +647,12 @@ bool        _CalcNode::RecomputeMatrix  (long categID, long totalCategs, _Matrix
     }
     return false;
 }
+//_______________________________________________________________________________________________
+
+long        _CalcNode::map_global_to_local_category (long cat) const {
+    if (remapMyCategories.lLength) return remapMyCategories.list_data[cat*(categoryVariables.lLength+1)];
+    return cat;
+}
 
 //_______________________________________________________________________________________________
 void        _CalcNode::SetCompExp  (_Matrix* m, long catID, bool do_exponentiation) {
@@ -738,7 +803,7 @@ void     _CalcNode::SetupCategoryMap (_List& containerVariables, _SimpleList& cl
         }
     }
     
-    //printf ("Node remap at %s yielded %s\n", GetName()->sData, _String((_String*)remapMyCategories.toStr()).sData);
+    //printf ("Node remap at %s yielded %s (%s)\n", GetName()->get_str(), _String((_String*)remapMyCategories.toStr()).get_str(),  _String((_String*)classCounter.toStr()).get_str());
     
 }
 
@@ -756,10 +821,17 @@ node<long>* _CalcNode::LocateMeInTree (void) const {
 
 //_______________________________________________________________________________________________
 
-void _CalcNode::ConvertToSimpleMatrix (void) const {
+void _CalcNode::ConvertToSimpleMatrix (unsigned long category_count) {
     _Formula * mf = GetExplicitFormModel();
     if (mf) {
-        mf->ConvertMatrixArgumentsToSimpleOrComplexForm (false);
+         if (!templateFormulaClone) {
+            templateFormulaClone = new _Formula* [category_count];
+            for (long i = 0; i < category_count; i++) {
+                templateFormulaClone[i] = new _Formula (*mf);
+                templateFormulaClone[i]->ConvertMatrixArgumentsToSimpleOrComplexForm (false);
+            }
+        }
+        
     } else {
         _Matrix * mm [2] = {GetModelMatrix(), GetFreqMatrix()};
         for (_Matrix * m : mm) {
@@ -772,10 +844,20 @@ void _CalcNode::ConvertToSimpleMatrix (void) const {
 
 //_______________________________________________________________________________________________
 
-void _CalcNode::ConvertFromSimpleMatrix (void) {
+void _CalcNode::ConvertFromSimpleMatrix (unsigned long category_count) {
     _Formula * mf = GetExplicitFormModel();
-    if (mf) {
-        mf->ConvertMatrixArgumentsToSimpleOrComplexForm (true);
+    if (templateFormulaClone || mf) {
+        if (templateFormulaClone) {
+            //printf ("_CalcNode::ConvertFromSimpleMatrix %s => %d\n", GetName()->get_str(), category_count);
+            for (long i = 0; i < category_count; i++) {
+                delete templateFormulaClone[i];
+            }
+            delete [] templateFormulaClone;
+            templateFormulaClone = nil;
+            GetExplicitFormModel()->ConvertMatrixArgumentsToSimpleOrComplexForm(true);
+        } else {
+            mf->ConvertMatrixArgumentsToSimpleOrComplexForm(true);
+        }
     } else {
         _Matrix * mm [2] = {GetModelMatrix(), GetFreqMatrix()};
         for (_Matrix * m : mm) {
